@@ -40,6 +40,8 @@
 
 guint _gdk_keymap_serial = 0;
 gboolean _gdk_keyboard_has_altgr = FALSE;
+guint _scancode_rshift = 0;
+
 static GdkModifierType gdk_shift_modifiers = GDK_SHIFT_MASK;
 
 static GdkKeymap *default_keymap = NULL;
@@ -142,6 +144,8 @@ handle_special (guint  vk,
       *ksymp = GDK_Meta_L; break;
     case VK_RWIN:
       *ksymp = GDK_Meta_R; break;
+    case VK_APPS:
+      *ksymp = GDK_Menu; break;
     case VK_MULTIPLY:
       *ksymp = GDK_KP_Multiply; break;
     case VK_ADD:
@@ -265,7 +269,7 @@ reset_after_dead (guchar key_state[256])
     }
 }
 
-static gboolean
+static void
 handle_dead (guint  keysym,
 	     guint *ksymp)
 {
@@ -306,9 +310,12 @@ handle_dead (guint  keysym,
     case GDK_Greek_accentdieresis: /* 0x7ae */
       *ksymp = GDK_Greek_accentdieresis; break;
     default:
-      return FALSE;
+      /* By default use the keysym as such. This takes care of for
+       * instance the dead U+09CD (BENGALI VIRAMA) on the ekushey
+       * Bengali layout.
+       */
+      *ksymp = keysym; break;
     }
-  return TRUE;
 }
 
 static void
@@ -357,6 +364,9 @@ update_keymap (void)
 	{
 	  gint shift;
 
+	  if (vk == VK_RSHIFT)
+	    _scancode_rshift = scancode;
+
 	  key_state[vk] = 0x80;
 	  for (shift = 0; shift < 4; shift++)
 	    {
@@ -446,14 +456,7 @@ update_keymap (void)
 		      reset_after_dead (key_state);
 
 		      /* Use dead keysyms instead of "undead" ones */
-		      if (!handle_dead (keysym, ksymp))
-			GDK_NOTE (EVENTS,
-				  g_print ("Unhandled dead key cp:%d vk:%02x sc:%x ch:%02x wc:%04x keysym:%04x%s%s\n",
-					   _gdk_input_codepage, vk,
-					   scancode, chars[0],
-					   wcs[0], keysym,
-					   (shift&0x1 ? " shift" : ""),
-					   (shift&0x2 ? " altgr" : "")));
+		      handle_dead (keysym, ksymp);
 		    }
 		  else if (k == 0)
 		    {
@@ -892,7 +895,7 @@ gdk_keymap_translate_keyboard_state (GdkKeymap       *keymap,
     }
 				
 #if 0
-  GDK_NOTE (EVENTS, g_print ("...group=%d level=%d cmods=%#x keyval=%s\n",
+  GDK_NOTE (EVENTS, g_print ("... group=%d level=%d cmods=%#x keyval=%s\n",
 			     group, shift_level, tmp_modifiers, gdk_keyval_name (tmp_keyval)));
 #endif
 
