@@ -26,6 +26,7 @@
 
 #include <config.h>
 
+#include <glib.h>
 #include "gdkconfig.h"
 
 #include <locale.h>
@@ -37,20 +38,19 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <gmodule.h>
-#ifdef G_OS_UNIX
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#include <sys/types.h>		/* For uid_t, gid_t */
 #endif
+#include <sys/types.h>		/* For uid_t, gid_t */
+
 #ifdef G_OS_WIN32
 #define STRICT
 #include <windows.h>
 #undef STRICT
 #endif
 
-#include <pango/pango-utils.h>	/* For pango_split_file_list */
+#include <pango/pango-types.h>	/* For pango_language_from_string */
 
-#include "gtkalias.h"
 #include "gtkintl.h"
 
 #include "gtkaccelmap.h"
@@ -59,6 +59,7 @@
 #include "gtkdnd.h"
 #include "gtkversion.h"
 #include "gtkmain.h"
+#include "gtkmodules.h"
 #include "gtkrc.h"
 #include "gtkselection.h"
 #include "gtksettings.h"
@@ -66,6 +67,7 @@
 #include "gtkwindow.h"
 #include "gtkprivate.h"
 #include "gtkdebug.h"
+#include "gtkalias.h"
 
 /* Private type definitions
  */
@@ -73,7 +75,6 @@ typedef struct _GtkInitFunction		 GtkInitFunction;
 typedef struct _GtkQuitFunction		 GtkQuitFunction;
 typedef struct _GtkClosure	         GtkClosure;
 typedef struct _GtkKeySnooperData	 GtkKeySnooperData;
-typedef struct _GtkModuleInfo            GtkModuleInfo;
 
 struct _GtkInitFunction
 {
@@ -105,12 +106,6 @@ struct _GtkKeySnooperData
   guint id;
 };
 
-struct _GtkModuleInfo
-{
-  GtkModuleInitFunc init_func;
-  GtkModuleDisplayInitFunc display_init_func;
-};
-
 static gint  gtk_quit_invoke_function	 (GtkQuitFunction    *quitf);
 static void  gtk_quit_destroy		 (GtkQuitFunction    *quitf);
 static gint  gtk_invoke_key_snoopers	 (GtkWidget	     *grab_widget,
@@ -137,13 +132,6 @@ const guint gtk_micro_version = GTK_MICRO_VERSION;
 const guint gtk_binary_age = GTK_BINARY_AGE;
 const guint gtk_interface_age = GTK_INTERFACE_AGE;
 
-static GSList *gtk_modules;
-
-/* Saved argc,argv for delayed module initialization
- */
-static gint    gtk_argc = 0;
-static gchar **gtk_argv = NULL;
-
 static guint gtk_main_loop_level = 0;
 static gint gtk_initialized = FALSE;
 static GList *current_events = NULL;
@@ -168,7 +156,10 @@ static const GDebugKey gtk_debug_keys[] = {
   {"tree", GTK_DEBUG_TREE},
   {"updates", GTK_DEBUG_UPDATES},
   {"keybindings", GTK_DEBUG_KEYBINDINGS},
-  {"multihead", GTK_DEBUG_MULTIHEAD}
+  {"multihead", GTK_DEBUG_MULTIHEAD},
+  {"modules", GTK_DEBUG_MODULES},
+  {"geometry", GTK_DEBUG_GEOMETRY},
+  {"icontheme", GTK_DEBUG_ICONTHEME}
 };
 
 static const guint gtk_ndebug_keys = sizeof (gtk_debug_keys) / sizeof (GDebugKey);
@@ -334,266 +325,13 @@ _gtk_get_data_prefix (void)
 
 #endif /* G_OS_WIN32 */
 
-static gchar **
-get_module_path (void)
-{
-  const gchar *module_path_env;
-  const gchar *exe_prefix;
-  const gchar *home_dir;
-  gchar *home_gtk_dir = NULL;
-  gchar *module_path;
-  gchar *default_dir;
-  static gchar **result = NULL;
-
-  if (result)
-    return result;
-
-  home_dir = g_get_home_dir();
-  if (home_dir)
-    home_gtk_dir = g_build_filename (home_dir, ".gtk-2.0", NULL);
-
-  module_path_env = g_getenv ("GTK_PATH");
-  exe_prefix = g_getenv ("GTK_EXE_PREFIX");
-
-  if (exe_prefix)
-    default_dir = g_build_filename (exe_prefix, "lib", "gtk-2.0", NULL);
-  else
-    default_dir = g_build_filename (GTK_LIBDIR, "gtk-2.0", NULL);
-
-  if (module_path_env && home_gtk_dir)
-    module_path = g_build_path (G_SEARCHPATH_SEPARATOR_S,
-				module_path_env, home_gtk_dir, default_dir, NULL);
-  else if (module_path_env)
-    module_path = g_build_path (G_SEARCHPATH_SEPARATOR_S,
-				module_path_env, default_dir, NULL);
-  else if (home_gtk_dir)
-    module_path = g_build_path (G_SEARCHPATH_SEPARATOR_S,
-				home_gtk_dir, default_dir, NULL);
-  else
-    module_path = g_build_path (G_SEARCHPATH_SEPARATOR_S,
-				default_dir, NULL);
-
-  g_free (home_gtk_dir);
-  g_free (default_dir);
-
-  result = pango_split_file_list (module_path);
-  g_free (module_path);
-
-  return result;
-}
-
-/**
- * _gtk_get_module_path:
- * @type: the type of the module, for instance 'modules', 'engines', immodules'
- * 
- * Determines the search path for a particular type of module.
- * 
- * Return value: the search path for the module type. Free with g_strfreev().
- **/
-gchar **
-_gtk_get_module_path (const gchar *type)
-{
-  gchar **paths = get_module_path();
-  gchar **path;
-  gchar **result;
-  gint count = 0;
-
-  for (path = paths; *path; path++)
-    count++;
-
-  result = g_new (gchar *, count * 4 + 1);
-
-  count = 0;
-  for (path = get_module_path (); *path; path++)
-    {
-      gint use_version, use_host;
-      
-      for (use_version = TRUE; use_version >= FALSE; use_version--)
-	for (use_host = TRUE; use_host >= FALSE; use_host--)
-	  {
-	    gchar *tmp_dir;
-	    
-	    if (use_version && use_host)
-	      tmp_dir = g_build_filename (*path, GTK_BINARY_VERSION, GTK_HOST, type, NULL);
-	    else if (use_version)
-	      tmp_dir = g_build_filename (*path, GTK_BINARY_VERSION, type, NULL);
-	    else if (use_host)
-	      tmp_dir = g_build_filename (*path, GTK_HOST, type, NULL);
-	    else
-	      tmp_dir = g_build_filename (*path, type, NULL);
-
-	    result[count++] = tmp_dir;
-	  }
-    }
-
-  result[count++] = NULL;
-
-  return result;
-}
-
-/* Like g_module_path, but use .la as the suffix
- */
-static gchar*
-module_build_la_path (const gchar *directory,
-		      const gchar *module_name)
-{
-	gchar *filename;
-	gchar *result;
-	
-	if (strncmp (module_name, "lib", 3) == 0)
-		filename = (gchar *)module_name;
-	else
-		filename =  g_strconcat ("lib", module_name, ".la", NULL);
-
-	if (directory && *directory)
-		result = g_build_filename (directory, filename, NULL);
-	else
-		result = g_strdup (filename);
-
-	if (filename != module_name)
-		g_free (filename);
-
-	return result;
-}
-
-/**
- * _gtk_find_module:
- * @name: the name of the module
- * @type: the type of the module, for instance 'modules', 'engines', immodules'
- * 
- * Looks for a dynamically loadable module named @name of type @type in the
- * standard GTK+ module search path.
- * 
- * Return value: the pathname to the found module, or %NULL if it wasn't found.
- *  Free with g_free().
- **/
-gchar *
-_gtk_find_module (const gchar *name,
-		  const gchar *type)
-{
-  gchar **paths;
-  gchar **path;
-  gchar *module_name = NULL;
-
-  if (g_path_is_absolute (name))
-    return g_strdup (name);
-
-  paths = _gtk_get_module_path (type);
-  for (path = paths; *path; path++)
-    {
-      gchar *tmp_name;
-
-      tmp_name = g_module_build_path (*path, name);
-      if (g_file_test (tmp_name, G_FILE_TEST_EXISTS))
-	{
-	  module_name = tmp_name;
-	  goto found;
-	}
-      g_free(tmp_name);
-
-      tmp_name = module_build_la_path (*path, name);
-      if (g_file_test (tmp_name, G_FILE_TEST_EXISTS))
-	{
-	  module_name = tmp_name;
-	  goto found;
-	}
-      g_free(tmp_name);
-    }
-
- found:
-  g_strfreev (paths);
-  return module_name;
-}
-
-static GModule *
-find_module (const gchar *name)
-{
-  GModule *module;
-  gchar *module_name;
-
-  module_name = _gtk_find_module (name, "modules");
-  if (!module_name)
-    {
-      /* As last resort, try loading without an absolute path (using system
-       * library path)
-       */
-      module_name = g_module_build_path (NULL, name);
-    }
-  
-  module = g_module_open (module_name, G_MODULE_BIND_LAZY);
-  g_free(module_name);
-
-  return module;
-}
-
-static GSList *
-load_module (GSList      *module_list,
-	     const gchar *name)
-{
-  GtkModuleInitFunc modinit_func = NULL;
-  GtkModuleInfo *info;
-  GModule *module = NULL;
-  
-  if (g_module_supported ())
-    {
-      module = find_module (name);
-      if (module &&
-	  g_module_symbol (module, "gtk_module_init", (gpointer *) &modinit_func) &&
-	  modinit_func)
-	{
-	  if (!g_slist_find (module_list, (gconstpointer) modinit_func))
-	    {
-	      g_module_make_resident (module);
-	      info = g_new (GtkModuleInfo, 1);
-
-	      info->init_func = modinit_func;
-	      g_module_symbol (module, "gtk_module_display_init",
-			       (gpointer *) &info->display_init_func);
-	      
-	      module_list = g_slist_prepend (module_list, info);
-	    }
-	  else
-	    {
-	      g_module_close (module);
-	      module = NULL;
-	    }
-	}
-    }
-  if (!modinit_func)
-    {
-      g_message ("Failed to load module \"%s\": %s",
-		 module ? g_module_name (module) : name,
-		 g_module_error ());
-      if (module)
-	g_module_close (module);
-    }
-  
-  return module_list;
-}
-
-static GSList *
-load_modules (const char *module_str)
-{
-  gchar **module_names = pango_split_file_list (module_str);
-  GSList *module_list = NULL;
-  gint i;
-  
-  for (i = 0; module_names[i]; i++)
-    module_list = load_module (module_list, module_names[i]);
-  
-  module_list = g_slist_reverse (module_list);
-  
-  g_strfreev (module_names);
-
-  return module_list;
-}
-
 static gboolean do_setlocale = TRUE;
 
 /**
  * gtk_disable_setlocale:
  * 
- * Prevents gtk_init() and gtk_init_check() from automatically
+ * Prevents gtk_init(), gtk_init_check(), gtk_init_with_args() and
+ * gtk_parse_args() from automatically
  * calling <literal>setlocale (LC_ALL, "")</literal>. You would 
  * want to use this function if you wanted to set the locale for 
  * your program to something other than the user's locale, or if 
@@ -614,51 +352,295 @@ gtk_disable_setlocale (void)
 #undef gtk_init_check
 #endif
 
-static void
-default_display_notify_cb (GdkDisplayManager *display_manager)
+static GString *gtk_modules_string = NULL;
+static gboolean g_fatal_warnings = FALSE;
+
+#ifdef G_ENABLE_DEBUG
+static gboolean
+gtk_arg_debug_cb (const char *key, const char *value, gpointer user_data)
 {
-  GSList *slist;
+  gtk_debug_flags |= g_parse_debug_string (value,
+					   gtk_debug_keys,
+					   gtk_ndebug_keys);
 
-  /* Initialize non-multihead-aware modules when the
-   * default display is first set to a non-NULL value.
-   */
-  static gboolean initialized = FALSE;
+  return TRUE;
+}
 
-  if (!gdk_display_get_default () || initialized)
+static gboolean
+gtk_arg_no_debug_cb (const char *key, const char *value, gpointer user_data)
+{
+  gtk_debug_flags &= ~g_parse_debug_string (value,
+					    gtk_debug_keys,
+					    gtk_ndebug_keys);
+
+  return TRUE;
+}
+#endif /* G_ENABLE_DEBUG */
+
+static gboolean
+gtk_arg_module_cb (const char *key, const char *value, gpointer user_data)
+{
+  if (value && *value)
+    {
+      if (gtk_modules_string)
+	g_string_append_c (gtk_modules_string, G_SEARCHPATH_SEPARATOR);
+      else
+	gtk_modules_string = g_string_new (NULL);
+      
+      g_string_append (gtk_modules_string, value);
+    }
+
+  return TRUE;
+}
+
+static GOptionEntry gtk_args[] = {
+  { "gtk-module",       0, 0, G_OPTION_ARG_CALLBACK, gtk_arg_module_cb,   
+    /* Description of --gtk-module=MODULES in --help output */ N_("Load additional GTK+ modules"), 
+    /* Placeholder in --gtk-module=MODULES in --help output */ N_("MODULES") },
+  { "g-fatal-warnings", 0, 0, G_OPTION_ARG_NONE, &g_fatal_warnings, 
+    /* Description of --g-fatal-warnings in --help output */   N_("Make all warnings fatal"), NULL },
+#ifdef G_ENABLE_DEBUG
+  { "gtk-debug",        0, 0, G_OPTION_ARG_CALLBACK, gtk_arg_debug_cb,    
+    /* Description of --gtk-debug=FLAGS in --help output */    N_("GTK+ debugging flags to set"), 
+    /* Placeholder in --gtk-debug=FLAGS in --help output */    N_("FLAGS") },
+  { "gtk-no-debug",     0, 0, G_OPTION_ARG_CALLBACK, gtk_arg_no_debug_cb, 
+    /* Description of --gtk-no-debug=FLAGS in --help output */ N_("GTK+ debugging flags to unset"), 
+    /* Placeholder in --gtk-no-debug=FLAGS in --help output */ N_("FLAGS") },
+#endif 
+  { NULL }
+};
+
+static void
+do_pre_parse_initialization (int    *argc,
+			     char ***argv)
+{
+  const gchar *env_string;
+  
+#if	0
+  g_set_error_handler (gtk_error);
+  g_set_warning_handler (gtk_warning);
+  g_set_message_handler (gtk_message);
+  g_set_print_handler (gtk_print);
+#endif
+
+  if (do_setlocale)
+    {
+      if (!setlocale (LC_ALL, ""))
+	g_warning ("Locale not supported by C library.\n\tUsing the fallback 'C' locale.");
+    }
+
+  gdk_pre_parse_libgtk_only ();
+  gdk_event_handler_set ((GdkEventFunc)gtk_main_do_event, NULL, NULL);
+  
+#ifdef G_ENABLE_DEBUG
+  env_string = g_getenv ("GTK_DEBUG");
+  if (env_string != NULL)
+    {
+      gtk_debug_flags = g_parse_debug_string (env_string,
+					      gtk_debug_keys,
+					      gtk_ndebug_keys);
+      env_string = NULL;
+    }
+#endif	/* G_ENABLE_DEBUG */
+
+  env_string = g_getenv ("GTK_MODULES");
+  if (env_string)
+    gtk_modules_string = g_string_new (env_string);
+}
+
+static void
+gettext_initialization (void)
+{
+#ifdef ENABLE_NLS
+  bindtextdomain (GETTEXT_PACKAGE, GTK_LOCALEDIR);
+  bindtextdomain (GETTEXT_PACKAGE "-properties", GTK_LOCALEDIR);
+#    ifdef HAVE_BIND_TEXTDOMAIN_CODESET
+  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+  bind_textdomain_codeset (GETTEXT_PACKAGE "-properties", "UTF-8");
+#    endif
+#endif  
+}
+
+static void
+do_post_parse_initialization (int    *argc,
+			      char ***argv)
+{
+  if (gtk_initialized)
     return;
 
-  initialized = TRUE;
+  gettext_initialization ();
 
-  for (slist = gtk_modules; slist; slist = slist->next)
+  if (g_fatal_warnings)
     {
-      if (slist->data)
-	{
-	  GtkModuleInfo *info = slist->data;
+      GLogLevelFlags fatal_mask;
 
-	  if (!info->display_init_func)
-	    info->init_func (&gtk_argc, &gtk_argv);
-	}
+      fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
+      fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
+      g_log_set_always_fatal (fatal_mask);
+    }
+
+  if (gtk_debug_flags & GTK_DEBUG_UPDATES)
+    gdk_window_set_debug_updates (TRUE);
+
+  {
+  /* Translate to default:RTL if you want your widgets
+   * to be RTL, otherwise translate to default:LTR.
+   * Do *not* translate it to "predefinito:LTR", if it
+   * it isn't default:LTR or default:RTL it will not work 
+   */
+    char *e = _("default:LTR");
+    if (strcmp (e, "default:RTL")==0) {
+      gtk_widget_set_default_direction (GTK_TEXT_DIR_RTL);
+    } else if (strcmp (e, "default:LTR")) {
+      g_warning ("Whoever translated default:LTR did so wrongly.\n");
+    }
+  }
+
+  gtk_type_init (0);
+  _gtk_accel_map_init ();  
+  _gtk_rc_init ();
+
+  /* Set the 'initialized' flag.
+   */
+  gtk_initialized = TRUE;
+
+  /* load gtk modules */
+  if (gtk_modules_string)
+    {
+      _gtk_modules_init (argc, argv, gtk_modules_string->str);
+      g_string_free (gtk_modules_string, TRUE);
     }
 }
 
-static void
-display_opened_cb (GdkDisplayManager *display_manager,
-		   GdkDisplay        *display)
+
+typedef struct
 {
-  GSList *slist;
-  
-  for (slist = gtk_modules; slist; slist = slist->next)
-    {
-      if (slist->data)
-	{
-	  GtkModuleInfo *info = slist->data;
+  gboolean open_default_display;
+} OptionGroupInfo;
 
-	  if (info->display_init_func)
-	    info->display_init_func (display);
-	}
-    }
+static gboolean
+pre_parse_hook (GOptionContext *context,
+		GOptionGroup   *group,
+		gpointer	data,
+		GError        **error)
+{
+  do_pre_parse_initialization (NULL, NULL);
+  
+  return TRUE;
 }
 
+static gboolean
+post_parse_hook (GOptionContext *context,
+		 GOptionGroup   *group,
+		 gpointer	data,
+		 GError        **error)
+{
+  OptionGroupInfo *info = data;
+
+  
+  do_post_parse_initialization (NULL, NULL);
+  
+  if (info->open_default_display)
+    return gdk_display_open_default_libgtk_only () != NULL;
+  else
+    return TRUE;
+}
+
+
+/**
+ * gtk_get_option_group:
+ * @open_default_display: whether to open the default display 
+ *    when parsing the commandline arguments
+ * 
+ * Returns a #GOptionGroup for the commandline arguments recognized
+ * by GTK+ and GDK. You should add this group to your #GOptionContext 
+ * with g_option_context_add_group(), if you are using 
+ * g_option_context_parse() to parse your commandline arguments.
+ *
+ * Returns a #GOptionGroup for the commandline arguments recognized
+ *   by GTK+
+ *
+ * Since: 2.6
+ */
+GOptionGroup *
+gtk_get_option_group (gboolean open_default_display)
+{
+  GOptionGroup *group;
+  OptionGroupInfo *info;
+
+  info = g_new0 (OptionGroupInfo, 1);
+  info->open_default_display = open_default_display;
+  
+  group = g_option_group_new ("gtk", _("GTK+ Options"), _("Show GTK+ Options"), info, g_free);
+  g_option_group_set_parse_hooks (group, pre_parse_hook, post_parse_hook);
+
+  gdk_add_option_entries_libgtk_only (group);
+  g_option_group_add_entries (group, gtk_args);
+  g_option_group_set_translation_domain (group, GETTEXT_PACKAGE);
+  
+  return group;
+}
+
+/**
+ * gtk_init_with_args:
+ * @argc: a pointer to the number of command line arguments.
+ * @argv: a pointer to the array of command line arguments.
+ * @parameter_string: a string which is displayed in
+ *    the first line of <option>--help</option> output, after 
+ *    <literal><replaceable>programname</replaceable> [OPTION...]</literal>
+ * @entries: a %NULL-terminated array of #GOptionEntry<!-- -->s
+ *    describing the options of your program
+ * @translation_domain: a translation domain to use for translating
+ *    the <option>--help</option> output for the options in @entries
+ *    with gettext(), or %NULL
+ * @error: a return location for errors 
+ *
+ * This function does the same work as gtk_init_check(). 
+ * Additionally, it allows you to add your own commandline options, 
+ * and it automatically generates nicely formatted 
+ * <option>--help</option> output. Note that your program will
+ * be terminated after writing out the help output.
+ *
+ * Returns: %TRUE if the GUI has been successfully initialized, 
+ *               %FALSE otherwise.
+ * 
+ * Since: 2.6
+ */
+gboolean
+gtk_init_with_args (int            *argc,
+		    char         ***argv,
+		    char           *parameter_string,  
+		    GOptionEntry   *entries,
+		    char           *translation_domain,
+		    GError        **error)
+{
+  GOptionContext *context;
+  GOptionGroup *gtk_group;
+  gboolean retval;
+
+  if (gtk_initialized)
+    return TRUE;
+
+  gettext_initialization ();
+
+  if (!check_setugid ())
+    return FALSE;
+
+  gtk_group = gtk_get_option_group (TRUE);
+  
+  context = g_option_context_new (parameter_string);
+  g_option_context_add_group (context, gtk_group);
+  
+  if (entries)
+    g_option_context_add_main_entries (context, entries, translation_domain);
+  retval = g_option_context_parse (context, argc, argv, error);
+  
+  g_option_context_free (context);
+
+  return retval;
+}
+
+
 /**
  * gtk_parse_args:
  * @argc: a pointer to the number of command line arguments.
@@ -680,217 +662,25 @@ gboolean
 gtk_parse_args (int    *argc,
 		char ***argv)
 {
-  GString *gtk_modules_string = NULL;
-  GSList *slist;
-  GdkDisplayManager *display_manager;
-  const gchar *env_string;
-
+  GOptionContext *option_context;
+  GOptionGroup *gtk_group;
+  
   if (gtk_initialized)
     return TRUE;
 
+  gettext_initialization ();
+
   if (!check_setugid ())
     return FALSE;
-  
-#if	0
-  g_set_error_handler (gtk_error);
-  g_set_warning_handler (gtk_warning);
-  g_set_message_handler (gtk_message);
-  g_set_print_handler (gtk_print);
-#endif
 
-  if (do_setlocale)
-    {
-      if (!setlocale (LC_ALL, ""))
-	g_warning ("Locale not supported by C library.\n\tUsing the fallback 'C' locale.");
-    }
+  option_context = g_option_context_new (NULL);
+  g_option_context_set_ignore_unknown_options (option_context, TRUE);
+  g_option_context_set_help_enabled (option_context, FALSE);
+  gtk_group = gtk_get_option_group (FALSE);
+  g_option_context_set_main_group (option_context, gtk_group);
+  g_option_context_parse (option_context, argc, argv, NULL);
+  g_option_context_free (option_context);
 
-  gdk_parse_args (argc, argv);
-  gdk_event_handler_set ((GdkEventFunc)gtk_main_do_event, NULL, NULL);
-  
-#ifdef G_ENABLE_DEBUG
-  env_string = g_getenv ("GTK_DEBUG");
-  if (env_string != NULL)
-    {
-      gtk_debug_flags = g_parse_debug_string (env_string,
-					      gtk_debug_keys,
-					      gtk_ndebug_keys);
-      env_string = NULL;
-    }
-#endif	/* G_ENABLE_DEBUG */
-
-  env_string = g_getenv ("GTK_MODULES");
-  if (env_string)
-    gtk_modules_string = g_string_new (env_string);
-
-  if (argc && argv)
-    {
-      gint i, j, k;
-      
-      for (i = 1; i < *argc;)
-	{
-	  if (strcmp ("--gtk-module", (*argv)[i]) == 0 ||
-	      strncmp ("--gtk-module=", (*argv)[i], 13) == 0)
-	    {
-	      gchar *module_name = (*argv)[i] + 12;
-	      
-	      if (*module_name == '=')
-		module_name++;
-	      else if (i + 1 < *argc)
-		{
-		  (*argv)[i] = NULL;
-		  i += 1;
-		  module_name = (*argv)[i];
-		}
-	      (*argv)[i] = NULL;
-
-	      if (module_name && *module_name)
-		{
-		  if (gtk_modules_string)
-		    g_string_append_c (gtk_modules_string, G_SEARCHPATH_SEPARATOR);
-		  else
-		    gtk_modules_string = g_string_new (NULL);
-
-		  g_string_append (gtk_modules_string, module_name);
-		}
-	    }
-	  else if (strcmp ("--g-fatal-warnings", (*argv)[i]) == 0)
-	    {
-	      GLogLevelFlags fatal_mask;
-	      
-	      fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
-	      fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
-              g_log_set_always_fatal (fatal_mask);
-	      (*argv)[i] = NULL;
-	    }
-#ifdef G_ENABLE_DEBUG
-	  else if ((strcmp ("--gtk-debug", (*argv)[i]) == 0) ||
-		   (strncmp ("--gtk-debug=", (*argv)[i], 12) == 0))
-	    {
-	      gchar *equal_pos = strchr ((*argv)[i], '=');
-	      
-	      if (equal_pos != NULL)
-		{
-		  gtk_debug_flags |= g_parse_debug_string (equal_pos+1,
-							   gtk_debug_keys,
-							   gtk_ndebug_keys);
-		}
-	      else if ((i + 1) < *argc && (*argv)[i + 1])
-		{
-		  gtk_debug_flags |= g_parse_debug_string ((*argv)[i+1],
-							   gtk_debug_keys,
-							   gtk_ndebug_keys);
-		  (*argv)[i] = NULL;
-		  i += 1;
-		}
-	      (*argv)[i] = NULL;
-	    }
-	  else if ((strcmp ("--gtk-no-debug", (*argv)[i]) == 0) ||
-		   (strncmp ("--gtk-no-debug=", (*argv)[i], 15) == 0))
-	    {
-	      gchar *equal_pos = strchr ((*argv)[i], '=');
-	      
-	      if (equal_pos != NULL)
-		{
-		  gtk_debug_flags &= ~g_parse_debug_string (equal_pos+1,
-							    gtk_debug_keys,
-							    gtk_ndebug_keys);
-		}
-	      else if ((i + 1) < *argc && (*argv)[i + 1])
-		{
-		  gtk_debug_flags &= ~g_parse_debug_string ((*argv)[i+1],
-							    gtk_debug_keys,
-							    gtk_ndebug_keys);
-		  (*argv)[i] = NULL;
-		  i += 1;
-		}
-	      (*argv)[i] = NULL;
-	    }
-#endif /* G_ENABLE_DEBUG */
-	  i += 1;
-	}
-      
-      for (i = 1; i < *argc; i++)
-	{
-	  for (k = i; k < *argc; k++)
-	    if ((*argv)[k] != NULL)
-	      break;
-	  
-	  if (k > i)
-	    {
-	      k -= i;
-	      for (j = i + k; j < *argc; j++)
-		(*argv)[j-k] = (*argv)[j];
-	      *argc -= k;
-	    }
-	}
-
-      gtk_argv = g_malloc ((gtk_argc + 1) * sizeof (char*));
-      for (i = 0; i < gtk_argc; i++)
-	gtk_argv[i] = g_strdup ((*argv)[i]);
-      gtk_argv[gtk_argc] = NULL;
-    }
-
-  if (gtk_debug_flags & GTK_DEBUG_UPDATES)
-    gdk_window_set_debug_updates (TRUE);
-
-  /* load gtk modules */
-  if (gtk_modules_string)
-    {
-      gtk_modules = load_modules (gtk_modules_string->str);
-      g_string_free (gtk_modules_string, TRUE);
-    }
-
-#ifdef ENABLE_NLS
-  bindtextdomain (GETTEXT_PACKAGE, GTK_LOCALEDIR);
-  bindtextdomain (GETTEXT_PACKAGE "-properties", GTK_LOCALEDIR);
-#    ifdef HAVE_BIND_TEXTDOMAIN_CODESET
-  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
-  bind_textdomain_codeset (GETTEXT_PACKAGE "-properties", "UTF-8");
-#    endif
-#endif  
-
-  {
-  /* Translate to default:RTL if you want your widgets
-   * to be RTL, otherwise translate to default:LTR.
-   * Do *not* translate it to "predefinito:LTR", if it
-   * it isn't default:LTR or default:RTL it will not work 
-   */
-    char *e = _("default:LTR");
-    if (strcmp (e, "default:RTL")==0) {
-      gtk_widget_set_default_direction (GTK_TEXT_DIR_RTL);
-    } else if (strcmp (e, "default:LTR")) {
-      g_warning ("Whoever translated default:LTR did so wrongly.\n");
-    }
-  }
-
-  gtk_type_init (0);
-  _gtk_accel_map_init ();  
-  _gtk_rc_init ();
-  
-  /* Set the 'initialized' flag.
-   */
-  gtk_initialized = TRUE;
-
-  display_manager = gdk_display_manager_get ();
-  g_signal_connect (display_manager, "notify::default-display",
-		    G_CALLBACK (default_display_notify_cb), NULL);
-  g_signal_connect (display_manager, "display-opened",
-		    G_CALLBACK (display_opened_cb), NULL);
-
-  /* initialize multhead aware gtk modules; for other modules,
-   * we wait until we have a display open;
-   */
-  for (slist = gtk_modules; slist; slist = slist->next)
-    {
-      if (slist->data)
-	{
-	  GtkModuleInfo *info = slist->data;
-
-	  if (info->display_init_func)
-	    info->init_func (argc, argv);
-	}
-    }
-  
   return TRUE;
 }
 
@@ -940,7 +730,13 @@ gtk_init_check (int	 *argc,
  * applications.  It will initialize everything needed to operate the 
  * toolkit and parses some standard command line options. @argc and 
  * @argv are adjusted accordingly so your own code will 
- * never see those standard arguments.
+ * never see those standard arguments. 
+ *
+ * Note that there are some alternative ways to initialize GTK+: 
+ * if you are calling gtk_parse_args(), gtk_init_check(), 
+ * gtk_init_with_args() or g_option_context_parse() with 
+ * the option group returned by gtk_get_option_group(), you 
+ * <emphasis>don't</emphasis> have to call gtk_init().
  *
  * <note><para>
  * This function will terminate your program if it was unable to initialize 
@@ -969,9 +765,8 @@ check_sizeof_GtkWindow (size_t sizeof_GtkWindow)
 	     "The code using GTK+ thinks GtkWindow is of different\n"
              "size than it actually is in this build of GTK+.\n"
 	     "On Windows, this probably means that you have compiled\n"
-	     "your code with gcc without the -fnative-struct\n"
-	     "(or -mms-bitfields) switch, or that you are using\n"
-	     "an unsupported compiler.");
+	     "your code with gcc without the -mms-bitfields switch,\n"
+	     "or that you are using an unsupported compiler.");
 }
 
 /* In GTK+ 2.0 the GtkWindow struct actually is the same size in
@@ -988,9 +783,8 @@ check_sizeof_GtkBox (size_t sizeof_GtkBox)
 	     "The code using GTK+ thinks GtkBox is of different\n"
              "size than it actually is in this build of GTK+.\n"
 	     "On Windows, this probably means that you have compiled\n"
-	     "your code with gcc without the -fnative-struct\n"
-	     "(or -mms-bitfields) switch, or that you are using\n"
-	     "an unsupported compiler.");
+	     "your code with gcc without the -mms-bitfields switch,\n"
+	     "or that you are using an unsupported compiler.");
 }
 
 /* These two functions might get more checks added later, thus pass
@@ -1218,12 +1012,16 @@ gtk_main (void)
 
       gdk_flush ();
     }
-	      
+    
   main_loops = g_slist_remove (main_loops, loop);
 
   g_main_loop_unref (loop);
 
   gtk_main_loop_level--;
+
+  /* Try storing all clipboard data we have */
+  if (gtk_main_loop_level == 0)
+    _gtk_clipboard_store_all ();
 }
 
 guint
@@ -1571,6 +1369,7 @@ gtk_main_do_event (GdkEvent *event)
     case GDK_CLIENT_EVENT:
     case GDK_VISIBILITY_NOTIFY:
     case GDK_WINDOW_STATE:
+    case GDK_GRAB_BROKEN:
       gtk_widget_event (event_widget, event);
       break;
 
@@ -2500,3 +2299,6 @@ _gtk_boolean_handled_accumulator (GSignalInvocationHint *ihint,
   
   return continue_emission;
 }
+
+#define __GTK_MAIN_C__
+#include "gtkaliasdef.c"
