@@ -31,17 +31,19 @@
 #include <config.h>
 
 #include <string.h>
-#include "gtkalias.h"
 #include "gtkintl.h"
 #include "gtkmarshalers.h"
 #include "gtkmenu.h"
 #include "gtkmenubar.h"
 #include "gtkmenushell.h"
+#include "gtkmenutoolbutton.h"
 #include "gtkseparatormenuitem.h"
 #include "gtkseparatortoolitem.h"
 #include "gtktearoffmenuitem.h"
 #include "gtktoolbar.h"
 #include "gtkuimanager.h"
+#include "gtkprivate.h"
+#include "gtkalias.h"
 
 #undef DEBUG_UI_MANAGER
 
@@ -76,6 +78,7 @@ struct _Node {
   GList *uifiles;
 
   guint dirty : 1;
+  guint expand : 1;  /* used for separators */
 };
 
 #define GTK_UI_MANAGER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_UI_MANAGER, GtkUIManagerPrivate))
@@ -124,6 +127,7 @@ static void        dirty_all_nodes                (GtkUIManager      *self);
 static void        mark_node_dirty                (GNode             *node);
 static GNode     * get_child_node                 (GtkUIManager      *self,
                                                    GNode             *parent,
+						   GNode             *sibling,
                                                    const gchar       *childname,
                                                    gint               childname_length,
                                                    NodeType           node_type,
@@ -224,11 +228,11 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    */
   g_object_class_install_property (gobject_class,
                                    PROP_ADD_TEAROFFS,
-                                   g_param_spec_boolean ("add_tearoffs",
+                                   g_param_spec_boolean ("add-tearoffs",
 							 P_("Add tearoffs to menus"),
 							 P_("Whether tearoff menu items should be added to menus"),
                                                          FALSE,
-							 G_PARAM_READWRITE));
+							 GTK_PARAM_READWRITE));
 
   g_object_class_install_property (gobject_class,
 				   PROP_UI,
@@ -236,7 +240,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
  							P_("Merged UI definition"),
 							P_("An XML string describing the merged UI"),
 							NULL,
-							G_PARAM_READABLE));
+							GTK_PARAM_READABLE));
 
 
   /**
@@ -400,7 +404,7 @@ gtk_ui_manager_init (GtkUIManager *self)
   self->private_data->add_tearoffs = FALSE;
 
   merge_id = gtk_ui_manager_new_merge_id (self);
-  node = get_child_node (self, NULL, "ui", 2,
+  node = get_child_node (self, NULL, NULL, "ui", 2,
 			 NODE_TYPE_ROOT, TRUE, FALSE);
   node_prepend_ui_reference (node, merge_id, 0);
 }
@@ -574,7 +578,7 @@ gtk_ui_manager_set_add_tearoffs (GtkUIManager *self,
       
       dirty_all_nodes (self);
 
-      g_object_notify (G_OBJECT (self), "add_tearoffs");
+      g_object_notify (G_OBJECT (self), "add-tearoffs");
     }
 }
 
@@ -736,6 +740,11 @@ gtk_ui_manager_get_accel_group (GtkUIManager *self)
  *
  * Note that the widget found by following a path that ends in a &lt;menu&gt;
  * element is the menuitem to which the menu is attached, not the menu itself.
+ *
+ * Also note that the widgets constructed by a ui manager are not tied to 
+ * the lifecycle of the ui manager. If you add the widgets returned by this 
+ * function to some container or explicitly ref them, they will survive the
+ * destruction of the ui manager.
  * 
  * Return value: the widget found by following the path, or %NULL if no widget
  *   was found.
@@ -844,6 +853,7 @@ gtk_ui_manager_get_action (GtkUIManager *self,
 static GNode *
 get_child_node (GtkUIManager *self, 
 		GNode        *parent,
+		GNode        *sibling,
 		const gchar  *childname, 
 		gint          childname_length,
 		NodeType      node_type,
@@ -852,11 +862,6 @@ get_child_node (GtkUIManager *self,
 {
   GNode *child = NULL;
 
-  g_return_val_if_fail (parent == NULL ||
-			(NODE_INFO (parent)->type != NODE_TYPE_MENUITEM &&
-			 NODE_INFO (parent)->type != NODE_TYPE_TOOLITEM), 
-			NULL);
-
   if (parent)
     {
       if (childname)
@@ -892,10 +897,22 @@ get_child_node (GtkUIManager *self,
 	  mnode->type = node_type;
 	  mnode->name = g_strndup (childname, childname_length);
 
-	  if (top)
-	    child = g_node_prepend_data (parent, mnode);
+	  if (sibling)
+	    {
+	      if (top)
+		child = g_node_insert_before (parent, sibling, 
+					      g_node_new (mnode));
+	      else
+		child = g_node_insert_after (parent, sibling, 
+					     g_node_new (mnode));
+	    }
 	  else
-	    child = g_node_append_data (parent, mnode);
+	    {
+	      if (top)
+		child = g_node_prepend_data (parent, mnode);
+	      else
+		child = g_node_append_data (parent, mnode);
+	    }
 
 	  mark_node_dirty (child);
 	}
@@ -954,7 +971,7 @@ get_node (GtkUIManager *self,
       else
 	length = strlen (pos);
 
-      node = get_child_node (self, parent, pos, length, NODE_TYPE_UNDECIDED,
+      node = get_child_node (self, parent, NULL, pos, length, NODE_TYPE_UNDECIDED,
 			     create, FALSE);
       if (!node)
 	return NULL;
@@ -1097,6 +1114,7 @@ start_element_handler (GMarkupParseContext *context,
   const gchar *action;
   GQuark action_quark;
   gboolean top;
+  gboolean expand = FALSE;
   
   gboolean raise_error = TRUE;
 
@@ -1120,6 +1138,10 @@ start_element_handler (GMarkupParseContext *context,
 	{
 	  top = !strcmp (attribute_values[i], "top");
 	}
+      else if (!strcmp (attribute_names[i], "expand"))
+	{
+	  expand = !strcmp (attribute_values[i], "true");
+	}
       else
 	{
 	  gint line_number, char_number;
@@ -1152,7 +1174,7 @@ start_element_handler (GMarkupParseContext *context,
       if (ctx->state == STATE_ROOT && !strcmp (element_name, "accelerator"))
 	{
 	  ctx->state = STATE_ACCELERATOR;
-	  ctx->current = get_child_node (self, ctx->current,
+	  ctx->current = get_child_node (self, ctx->current, NULL,
 					 node_name, strlen (node_name),
 					 NODE_TYPE_ACCELERATOR,
 					 TRUE, FALSE);
@@ -1178,7 +1200,7 @@ start_element_handler (GMarkupParseContext *context,
       if (ctx->state == STATE_ROOT && !strcmp (element_name, "menubar"))
 	{
 	  ctx->state = STATE_MENU;
-	  ctx->current = get_child_node (self, ctx->current,
+	  ctx->current = get_child_node (self, ctx->current, NULL,
 					 node_name, strlen (node_name),
 					 NODE_TYPE_MENUBAR,
 					 TRUE, FALSE);
@@ -1192,7 +1214,22 @@ start_element_handler (GMarkupParseContext *context,
 	}
       else if (ctx->state == STATE_MENU && !strcmp (element_name, "menu"))
 	{
-	  ctx->current = get_child_node (self, ctx->current,
+	  ctx->current = get_child_node (self, ctx->current, NULL,
+					 node_name, strlen (node_name),
+					 NODE_TYPE_MENU,
+					 TRUE, top);
+	  if (NODE_INFO (ctx->current)->action_name == 0)
+	    NODE_INFO (ctx->current)->action_name = action_quark;
+
+	  node_prepend_ui_reference (ctx->current, ctx->merge_id, action_quark);
+	  
+	  raise_error = FALSE;
+	}
+      else if (ctx->state == STATE_TOOLITEM && !strcmp (element_name, "menu"))
+	{
+	  ctx->state = STATE_MENU;
+	  
+	  ctx->current = get_child_node (self, g_node_last_child (ctx->current), NULL,
 					 node_name, strlen (node_name),
 					 NODE_TYPE_MENU,
 					 TRUE, top);
@@ -1208,7 +1245,7 @@ start_element_handler (GMarkupParseContext *context,
 	  GNode *node;
 
 	  ctx->state = STATE_MENUITEM;
-	  node = get_child_node (self, ctx->current,
+	  node = get_child_node (self, ctx->current, NULL,
 				 node_name, strlen (node_name),
 				 NODE_TYPE_MENUITEM,
 				 TRUE, top);
@@ -1224,7 +1261,7 @@ start_element_handler (GMarkupParseContext *context,
       if (ctx->state == STATE_ROOT && !strcmp (element_name, "popup"))
 	{
 	  ctx->state = STATE_MENU;
-	  ctx->current = get_child_node (self, ctx->current,
+	  ctx->current = get_child_node (self, ctx->current, NULL,
 					 node_name, strlen (node_name),
 					 NODE_TYPE_POPUP,
 					 TRUE, FALSE);
@@ -1239,12 +1276,12 @@ start_element_handler (GMarkupParseContext *context,
 	       !strcmp (element_name, "placeholder"))
 	{
 	  if (ctx->state == STATE_TOOLBAR)
-	    ctx->current = get_child_node (self, ctx->current,
+	    ctx->current = get_child_node (self, ctx->current, NULL,
 					   node_name, strlen (node_name),
 					   NODE_TYPE_TOOLBAR_PLACEHOLDER,
 					   TRUE, top);
 	  else
-	    ctx->current = get_child_node (self, ctx->current,
+	    ctx->current = get_child_node (self, ctx->current, NULL,
 					   node_name, strlen (node_name),
 					   NODE_TYPE_MENU_PLACEHOLDER,
 					   TRUE, top);
@@ -1272,11 +1309,13 @@ start_element_handler (GMarkupParseContext *context,
 	    }
 	  else
 	    length = strlen (node_name);
-	  node = get_child_node (self, ctx->current,
+	  node = get_child_node (self, ctx->current, NULL,
 				 node_name, length,
 				 NODE_TYPE_SEPARATOR,
 				 TRUE, top);
 
+	  NODE_INFO (node)->expand = expand;
+
 	  if (NODE_INFO (node)->action_name == 0)
 	    NODE_INFO (node)->action_name = action_quark;
 
@@ -1289,7 +1328,7 @@ start_element_handler (GMarkupParseContext *context,
       if (ctx->state == STATE_ROOT && !strcmp (element_name, "toolbar"))
 	{
 	  ctx->state = STATE_TOOLBAR;
-	  ctx->current = get_child_node (self, ctx->current,
+	  ctx->current = get_child_node (self, ctx->current, NULL,
 					 node_name, strlen (node_name),
 					 NODE_TYPE_TOOLBAR,
 					 TRUE, FALSE);
@@ -1305,7 +1344,7 @@ start_element_handler (GMarkupParseContext *context,
 	  GNode *node;
 
 	  ctx->state = STATE_TOOLITEM;
-	  node = get_child_node (self, ctx->current,
+	  node = get_child_node (self, ctx->current, NULL,
 				node_name, strlen (node_name),
 				 NODE_TYPE_TOOLITEM,
 				 TRUE, top);
@@ -1359,6 +1398,11 @@ end_element_handler (GMarkupParseContext *context,
       ctx->current = ctx->current->parent;
       if (NODE_INFO (ctx->current)->type == NODE_TYPE_ROOT) 
 	ctx->state = STATE_ROOT;
+      else if (NODE_INFO (ctx->current)->type == NODE_TYPE_TOOLITEM)
+	{
+	  ctx->current = ctx->current->parent;
+	  ctx->state = STATE_TOOLITEM;
+	}
       /* else, stay in same state */
       break;
     case STATE_MENUITEM:
@@ -1570,6 +1614,9 @@ gtk_ui_manager_add_ui_from_file (GtkUIManager *self,
  * separator if such an element can be inserted at the place determined by 
  * @path. Otherwise @type must indicate an element that can be inserted at 
  * the place determined by @path.
+ *
+ * If @path points to a menuitem or toolitem, the new element will be inserted
+ * before or after this item, depending on @top.
  * 
  * Since: 2.4
  **/
@@ -1583,6 +1630,7 @@ gtk_ui_manager_add_ui (GtkUIManager        *self,
 		       gboolean             top)
 {
   GNode *node;
+  GNode *sibling;
   GNode *child;
   NodeType node_type;
   GQuark action_quark = 0;
@@ -1592,14 +1640,22 @@ gtk_ui_manager_add_ui (GtkUIManager        *self,
   g_return_if_fail (name != NULL);
 
   node = get_node (self, path, NODE_TYPE_UNDECIDED, FALSE);
-  
+  sibling = NULL;
+
   if (node == NULL)
     return;
 
   node_type = NODE_TYPE_UNDECIDED;
 
+ reswitch:
   switch (NODE_INFO (node)->type) 
     {
+    case NODE_TYPE_SEPARATOR:
+    case NODE_TYPE_MENUITEM:
+    case NODE_TYPE_TOOLITEM:
+      sibling = node;
+      node = node->parent;
+      goto reswitch;
     case NODE_TYPE_MENUBAR:
     case NODE_TYPE_MENU:
     case NODE_TYPE_POPUP:
@@ -1675,9 +1731,13 @@ gtk_ui_manager_add_ui (GtkUIManager        *self,
     }
 
   if (node_type == NODE_TYPE_UNDECIDED)
-    return;
+    {
+      g_warning ("item type %d not suitable for adding at '%s'", 
+		 type, path);
+      return;
+    }
    
-  child = get_child_node (self, node,
+  child = get_child_node (self, node, sibling,
 			  name, strlen (name),
 			  node_type, TRUE, top);
 
@@ -2056,7 +2116,6 @@ update_node (GtkUIManager *self,
   Node *info;
   GNode *child;
   GtkAction *action;
-  gchar *tooltip;
   const gchar *action_name;
   NodeUIReference *ref;
   
@@ -2128,7 +2187,10 @@ update_node (GtkUIManager *self,
 	  GtkWidget *menu;
 	  GList *siblings;
 	  
-	  menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
+	  if (GTK_IS_MENU (info->proxy))
+	    menu = info->proxy;
+	  else
+	    menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
 	  siblings = gtk_container_get_children (GTK_CONTAINER (menu));
 	  if (siblings != NULL && GTK_IS_TEAROFF_MENU_ITEM (siblings->data))
 	    {
@@ -2174,12 +2236,16 @@ update_node (GtkUIManager *self,
 	if (info->proxy &&  
 	    G_OBJECT_TYPE (info->proxy) != GTK_ACTION_GET_CLASS (action)->menu_item_type)
 	  {
-	    prev_submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
-	    if (prev_submenu)
+	    if (GTK_IS_MENU_ITEM (info->proxy))
 	      {
-		g_object_ref (prev_submenu);
-		gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), NULL);
+		prev_submenu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
+		if (prev_submenu)
+		  {
+		    g_object_ref (prev_submenu);
+		    gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), NULL);
+		}
 	      }
+
 	    gtk_action_disconnect_proxy (info->action, info->proxy);
 	    gtk_container_remove (GTK_CONTAINER (info->proxy->parent),
 				  info->proxy);
@@ -2192,16 +2258,13 @@ update_node (GtkUIManager *self,
 	    GtkWidget *menushell;
 	    gint pos;
 	    
-	    if (find_menu_position (node, &menushell, &pos))
+	    if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM ||
+		find_menu_position (node, &menushell, &pos))
 	      {
 		GtkWidget *tearoff;
 		GtkWidget *filler;
 		
-		info->proxy = gtk_action_create_menu_item (action);
-		g_object_ref (info->proxy);
-		gtk_object_sink (GTK_OBJECT (info->proxy));
 		menu = gtk_menu_new ();
-		gtk_widget_set_name (info->proxy, info->name);
 		gtk_widget_set_name (menu, info->name);
 		tearoff = gtk_tearoff_menu_item_new ();
 		gtk_widget_set_no_show_all (tearoff, TRUE);
@@ -2213,10 +2276,27 @@ update_node (GtkUIManager *self,
 		gtk_widget_set_sensitive (filler, FALSE);
 		gtk_widget_set_no_show_all (filler, TRUE);
 		gtk_menu_shell_append (GTK_MENU_SHELL (menu), filler);
-		gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), menu);
-		gtk_menu_shell_insert (GTK_MENU_SHELL (menushell), info->proxy, pos);
-		g_signal_connect (info->proxy, "notify::visible",
-				  G_CALLBACK (update_smart_separators), NULL);
+
+		if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM)
+		  {
+		    info->proxy = menu;
+		    g_object_ref (info->proxy);
+		    gtk_object_sink (GTK_OBJECT (info->proxy));
+		    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (NODE_INFO (node->parent)->proxy),
+						   menu);
+		  }
+		else
+		  {
+		    info->proxy = gtk_action_create_menu_item (action);
+		    g_object_ref (info->proxy);
+		    gtk_object_sink (GTK_OBJECT (info->proxy));
+		    g_signal_connect (info->proxy, "notify::visible",
+				      G_CALLBACK (update_smart_separators), NULL);
+		    gtk_widget_set_name (info->proxy, info->name);
+		    
+		    gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), menu);
+		    gtk_menu_shell_insert (GTK_MENU_SHELL (menushell), info->proxy, pos);
+		  }
 	      }
 	  }
 	else
@@ -2228,7 +2308,11 @@ update_node (GtkUIManager *self,
 				       prev_submenu);
 	    g_object_unref (prev_submenu);
 	  }
-	menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
+	
+	if (GTK_IS_MENU (info->proxy))
+	  menu = info->proxy;
+	else
+	  menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
 	siblings = gtk_container_get_children (GTK_CONTAINER (menu));
 	if (siblings != NULL && GTK_IS_TEAROFF_MENU_ITEM (siblings->data))
 	  {
@@ -2409,7 +2493,7 @@ update_node (GtkUIManager *self,
       break;
     case NODE_TYPE_TOOLITEM:
       /* remove the proxy if it is of the wrong type ... */
-      if (info->proxy &&  
+      if (info->proxy && 
 	  G_OBJECT_TYPE (info->proxy) != GTK_ACTION_GET_CLASS (action)->toolbar_item_type)
 	{
 	  g_signal_handlers_disconnect_by_func (info->proxy,
@@ -2436,14 +2520,6 @@ update_node (GtkUIManager *self,
 	      
 	      gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
 				  GTK_TOOL_ITEM (info->proxy), pos);
-	      
-	      /* FIXME: we must trigger the notify::tooltip handler, since 
-	       * tooltips on toolitems can't be set before the toolitem 
-	       * is added to the toolbar.
-	       */
-	      g_object_get (action, "tooltip", &tooltip, NULL);
-	      g_object_set (action, "tooltip", tooltip, NULL);
-	      g_free (tooltip);
 	    }
 	}
       else
@@ -2453,6 +2529,13 @@ update_node (GtkUIManager *self,
 						NULL);
 	  gtk_action_connect_proxy (action, info->proxy);
 	}
+
+      /* FIXME: we must trigger the notify::tooltip handler, since 
+       * tooltips on toolitems can't be set before the toolitem 
+       * is added to the toolbar.
+       */
+      g_object_notify (G_OBJECT (action), "tooltip");
+
       g_signal_connect (info->proxy, "notify::visible",
 			G_CALLBACK (update_smart_separators), NULL);
       break;
@@ -2462,6 +2545,7 @@ update_node (GtkUIManager *self,
 	{
 	  GtkWidget *toolbar;
 	  gint pos;
+	  gint separator_mode;
 	  
 	  if (GTK_IS_SEPARATOR_TOOL_ITEM (info->proxy))
 	    {
@@ -2479,9 +2563,19 @@ update_node (GtkUIManager *self,
 	      g_object_ref (info->proxy);
 	      gtk_object_sink (GTK_OBJECT (info->proxy));
 	      gtk_widget_set_no_show_all (info->proxy, TRUE);
+	      if (info->expand)
+		{
+		  gtk_tool_item_set_expand (GTK_TOOL_ITEM (item), TRUE);
+		  gtk_separator_tool_item_set_draw
+		    (GTK_SEPARATOR_TOOL_ITEM (item), FALSE);
+		  separator_mode = SEPARATOR_MODE_VISIBLE;
+		}
+		else
+		  separator_mode = SEPARATOR_MODE_SMART;
+
 	      g_object_set_data (G_OBJECT (info->proxy),
 				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_SMART));
+				 GINT_TO_POINTER (separator_mode));
 	      gtk_widget_show (info->proxy);
 	    }
 	}
@@ -2538,9 +2632,11 @@ update_node (GtkUIManager *self,
   
   if (info->proxy) 
     {
-      if (info->type == NODE_TYPE_MENU) 
+      if (info->type == NODE_TYPE_MENU && GTK_IS_MENU_ITEM (info->proxy)) 
 	update_smart_separators (gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy)));
-      else if (info->type == NODE_TYPE_TOOLBAR || info->type == NODE_TYPE_POPUP) 
+      else if (info->type == NODE_TYPE_MENU || 
+	       info->type == NODE_TYPE_TOOLBAR || 
+	       info->type == NODE_TYPE_POPUP) 
 	update_smart_separators (info->proxy);
     }
   
@@ -2662,7 +2758,7 @@ mark_node_dirty (GNode *node)
     NODE_INFO (p)->dirty = TRUE;  
 }
 
-static const gchar *open_tag_format[] = {
+static const gchar *const open_tag_format[] = {
   "%*s<UNDECIDED",
   "%*s<ui",
   "%*s<menubar",
@@ -2677,7 +2773,7 @@ static const gchar *open_tag_format[] = {
   "%*s<accelerator"
 };
 
-static const gchar *close_tag_format[] = {
+static const gchar *const close_tag_format[] = {
   "%*s</UNDECIDED>\n",
   "%*s</ui>\n",
   "%*s</menubar>\n",
@@ -2775,3 +2871,6 @@ gtk_ui_manager_add_ui_from_file (GtkUIManager *self,
 }
 
 #endif
+
+#define __GTK_UI_MANAGER_C__
+#include "gtkaliasdef.c"
